
// $Id: GeometryDialog.cpp,v 1.36 2014/09/26 08:06:19 hupereir Exp $
/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "Anchor.h"
#include "AnimatedTreeView.h"
#include "ApplicationDataDialog.h"
#include "Debug.h"
#include "GeometryConfirmationDialog.h"
#include "GeometryApplication.h"
#include "GeometryDialog.h"
#include "GeometryDialog.moc"
#include "GridLayout.h"
#include "IconNames.h"
#include "IconEngine.h"
#include "InformationDialog.h"
#include "Singleton.h"
#include "XmlOptions.h"

#include <QApplication>
#include <QDesktopWidget>
#include <QGroupBox>
#include <QLabel>

#include <algorithm>

//___________________________________________________
GeometryDialog::GeometryDialog( QWidget* parent ):
CustomDialog( parent, OkButton|CancelButton ),
OptionWidgetList( this )
{

    Debug::Throw( "GeometryDialog::GeometryDialog.\n" );
    setOptionName( "GEOMETRY_DIALOG" );

    QGroupBox* group;
    mainLayout().addWidget( group = new QGroupBox( tr( "Applications" ), this ) );

    QHBoxLayout* hLayout( new QHBoxLayout() );
    hLayout->setMargin(5);
    hLayout->setSpacing(5);
    group->setLayout( hLayout );

    // list
    hLayout->addWidget( list_ = new AnimatedTreeView( group ) );
    list_->setModel( &model_ );
    list_->setSortingEnabled( false );
    list_->setOptionName( "APPLIST" );

    list_->setDragEnabled(true);
    list_->setAcceptDrops(true);

    connect( &model_, SIGNAL(itemOrderChanged(int,int)), SLOT(_reorder(int,int)) );

    // buttons
    QVBoxLayout* vLayout( new QVBoxLayout() );
    vLayout->setMargin(0);
    hLayout->addLayout( vLayout );

    QPushButton* button;
    vLayout->addWidget( button = new QPushButton( tr( "Sort" ), group ) );
    connect( button, SIGNAL(clicked()), SLOT(_sort()) );
    button->setIcon( IconEngine::get( IconNames::Sort ) );
    button->setToolTip( tr( "Sort applications based on alignment and current geometry" ) );

    vLayout->addWidget( edit_ = new QPushButton( tr( "Edit" ), group ) );
    connect( edit_, SIGNAL(clicked()), SLOT(_edit()) );
    edit_->setIcon( IconEngine::get( IconNames::Edit ) );
    edit_->setToolTip( tr( "Edit selected application geometry" ) );

    vLayout->addWidget( up_ = new QPushButton( tr( "Move Up" ), group ) );
    connect( up_, SIGNAL(clicked()), this, SLOT(_up()) );
    up_->setIcon( IconEngine::get( IconNames::Up ) );
    up_->setToolTip( tr( "Move up selected application" ) );

    vLayout->addWidget( down_ = new QPushButton( tr( "Move Down" ), group ) );
    connect( down_, SIGNAL(clicked()), this, SLOT(_down()) );
    down_->setIcon( IconEngine::get( IconNames::Down ) );
    down_->setToolTip( tr( "Move down selected application" ) );

    vLayout->addWidget( reload_ = new QPushButton( tr( "Reload" ), group ) );
    connect( reload_, SIGNAL(clicked()), this, SLOT(reload()) );
    reload_->setIcon( IconEngine::get( IconNames::Reload ) );
    reload_->setToolTip( tr( "Reload applications geometry" ) );

    QFrame* frame;
    vLayout->addWidget( frame = new QFrame( group ) );
    frame->setFrameStyle( QFrame::HLine );

    QPushButton* configurationButton;
    vLayout->addWidget( configurationButton = new QPushButton( tr( "Configure" ), group ) );
    connect(configurationButton, SIGNAL(clicked()), &static_cast<GeometryApplication*>(Singleton::get().application())->configurationAction(), SLOT(trigger()));
    configurationButton->setIcon( IconEngine::get( IconNames::Configure ) );
    configurationButton->setToolTip( tr( "Configure Transparency-settings" ) );

    vLayout->addStretch(1);

    // set connections
    connect( list_->selectionModel(), SIGNAL(selectionChanged(QItemSelection,QItemSelection)), SLOT(_updateButtons()) );
    connect( list_->selectionModel(), SIGNAL(currentChanged(QModelIndex,QModelIndex)), SLOT(_updateButtons()) );
    connect( list_, SIGNAL(activated(QModelIndex)), SLOT(_edit()) );

    mainLayout().addWidget( group = new QGroupBox( tr( "Automatic Placement" ), this ) );
    vLayout = new QVBoxLayout();
    vLayout->setMargin(5);
    vLayout->setSpacing(5);
    group->setLayout( vLayout );

    GridLayout* gridLayout = new GridLayout();
    gridLayout->setMargin(0);
    gridLayout->setSpacing(5);
    gridLayout->setMaxCount(2);
    vLayout->addLayout( gridLayout );

    gridLayout->setColumnAlignment( 0, Qt::AlignRight|Qt::AlignVCenter );

    QLabel* label;
    gridLayout->addWidget( label = new QLabel( tr( "Spacing:" ), this ) );
    gridLayout->addWidget( spacingSpinBox_ = new OptionSpinBox( this, "SETTINGS_SPACING", false ) );
    spacingSpinBox_->setSuffix( tr( "px" ) );
    spacingSpinBox_->setMinimum( -64 );
    spacingSpinBox_->setMaximum( 64 );
    label->setBuddy( spacingSpinBox_ );
    addOptionWidget( spacingSpinBox_ );

    gridLayout->addWidget( label = new QLabel( tr( "Offset:" ), this ) );
    gridLayout->addWidget( offsetSpinBox_ = new OptionSpinBox( this, "SETTINGS_OFFSET", false ) );
    offsetSpinBox_->setSuffix( tr( "px" ) );
    offsetSpinBox_->setMinimum( 0 );
    offsetSpinBox_->setMaximum( 64 );
    label->setBuddy( offsetSpinBox_ );
    addOptionWidget( offsetSpinBox_ );

    gridLayout->addWidget( label = new QLabel( tr( "Alignment:" ), this ) );
    gridLayout->addWidget( alignmentComboBox_ = new OptionComboBox( this, "SETTINGS_ALIGNMENT" ) );
    alignmentComboBox_->addItems( QStringList()
        << tr( "Horizontal" )
        << tr( "Vertical" ) );
    label->setBuddy( alignmentComboBox_ );
    addOptionWidget( alignmentComboBox_ );

    anchorWidget_ = new OptionAnchorWidget( this, "SETTINGS_ANCHOR", gridLayout );
    addOptionWidget( anchorWidget_ );
    gridLayout->setColumnStretch( 1, 1 );

    buttonLayout().insertWidget( 1, apply_ = new QPushButton( tr( "Apply" ), this ), 0,  Qt::AlignRight|Qt::AlignVCenter );
    apply_->setIcon( IconEngine::get( IconNames::DialogAccept ) );
    connect( apply_, SIGNAL(clicked()), SLOT(_apply()) );

    // update buttons
    _updateButtons();

    connect( alignmentComboBox_, SIGNAL(currentIndexChanged(int)), SLOT(_sort(int)) );

    connect( &okButton(), SIGNAL(clicked()), SLOT(_applyUnchecked()) );
    connect( &cancelButton(), SIGNAL(clicked()), SLOT(_restoreDefaults()) );

    connect( Singleton::get().application(), SIGNAL(configurationChanged()), SLOT(_updateConfiguration()) );
    connect( qApp, SIGNAL(aboutToQuit()), SLOT(_saveConfiguration()) );

}

//____________________________________________________
void GeometryDialog::add( ApplicationData data )
{
    Debug::Throw( "GeometryDialog::add.\n" );
    model_.add( data );
    _sort();
    list_->resizeColumns();
}

//____________________________________________________
void GeometryDialog::reload( void )
{
    Debug::Throw( "GeometryDialog::reload.\n" );
    model_.clear();
    backup_.clear();
    emit geometryRequested();

}

//______________________________________________________________________
void GeometryDialog::_updateButtons( void )
{
    Debug::Throw( "GeometryDialog::_updateButtons.\n" );
    ApplicationDataModel::List selection( model_.get( list_->selectionModel()->selectedRows() ) );
    bool has_selection( !selection.empty() );

    edit_->setEnabled( has_selection );
    up_->setEnabled( has_selection && !list_->selectionModel()->isSelected( model_.index( 0, 0 ) ) );
    down_->setEnabled( has_selection && !list_->selectionModel()->isSelected( model_.index( model_.rowCount()-1, 0 ) ) );
}

//____________________________________________________
void GeometryDialog::_edit( void )
{

    Debug::Throw( "GeometryDialog::_edit.\n" );

    // retrieve selection
    QModelIndex current( list_->selectionModel()->currentIndex() );
    Q_ASSERT( current.isValid() );

    ApplicationData currentData( model_.get( current ) );
    ApplicationDataDialog dialog( this, currentData );
    if( !dialog.centerOnWidget( parentWidget() ).exec() ) return;

    ApplicationData data( dialog.data() );
    if( data.geometry() == currentData.geometry() ) return;

    // propagate order
    data.setOrder( current.row() );
    emit geometryChanged( data );

    return;

}

//______________________________________________________________________
void GeometryDialog::_sort( int index )
{
    Debug::Throw() << "GeometryDialog::_sort - index: " << index << endl;
    ApplicationDataModel::List applications( model_.get() );
    if( index == 1 ) std::sort( applications.begin(), applications.end(), ApplicationData::VerticalFTor() );
    else std::sort( applications.begin(), applications.end(), ApplicationData::HorizontalFTor() );
    model_.set( applications );
}

//______________________________________________________________________
void GeometryDialog::_sort( void )
{ _sort( alignmentComboBox_->currentIndex() ); }

//_________________________________________________________
void GeometryDialog::_up( void )
{

    Debug::Throw( "GeometryDialog::_up.\n" );
    ApplicationDataModel::List selection( model_.get( list_->selectionModel()->selectedRows() ) );
    if( selection.empty() ) return;

    // retrieve selected indexes in list and store in model
    QModelIndexList selectedIndexes( list_->selectionModel()->selectedRows() );
    ApplicationDataModel::List selectedData( model_.get( selectedIndexes ) );

    ApplicationDataModel::List currentData( model_.get() );
    ApplicationDataModel::List newData;

    for( ApplicationDataModel::List::const_iterator iter = currentData.begin(); iter != currentData.end(); ++iter )
    {

        // check if new list is not empty, current index is selected and last index is not.
        // if yes, move.
        if(
            !( newData.empty() ||
            selectedIndexes.indexOf( model_.index( *iter ) ) == -1 ||
            selectedIndexes.indexOf( model_.index( newData.back() ) ) != -1
            ) )
        {
            ApplicationData last( newData.back() );
            newData.removeLast();
            newData << *iter;
            newData << last;
        } else newData << *iter;

    }

    model_.set( newData );

    // restore selection
    list_->selectionModel()->select( model_.index( selectedData.front() ),  QItemSelectionModel::Clear|QItemSelectionModel::Select|QItemSelectionModel::Rows );
    for( ApplicationDataModel::List::const_iterator iter = selectedData.begin(); iter != selectedData.end(); ++iter )
    { list_->selectionModel()->select( model_.index( *iter ), QItemSelectionModel::Select|QItemSelectionModel::Rows ); }

    return;

}

//_________________________________________________________
void GeometryDialog::_down( void )
{

    Debug::Throw( "GeometryDialog::_down.\n" );
    ApplicationDataModel::List selection( model_.get( list_->selectionModel()->selectedRows() ) );
    if( selection.empty() ) return;

    // retrieve selected indexes in list and store in model
    QModelIndexList selectedIndexes( list_->selectionModel()->selectedIndexes() );
    ApplicationDataModel::List selectedData( model_.get( selectedIndexes ) );

    ApplicationDataModel::List newData;
    ApplicationDataModel::ListIterator iter( model_.get() );
    iter.toBack();
    while( iter.hasPrevious() )
    {
        const ApplicationData& current( iter.previous() );

        // check if new list is not empty, current index is selected and last index is not.
        // if yes, move.
        if(
            !( newData.empty() ||
            selectedIndexes.indexOf( model_.index( current ) ) == -1 ||
            selectedIndexes.indexOf( model_.index( newData.front() ) ) != -1
            ) )
        {

            ApplicationData first( newData.front() );
            newData.removeFirst();
            newData.push_front( current );
            newData.push_front( first );

        } else newData.push_front( current );
    }

    model_.set( newData );

    // restore selection
    list_->selectionModel()->select( model_.index( selectedData.front() ),  QItemSelectionModel::Clear|QItemSelectionModel::Select|QItemSelectionModel::Rows );
    for( ApplicationDataModel::List::const_iterator iter = selectedData.begin(); iter != selectedData.end(); ++iter )
    { list_->selectionModel()->select( model_.index( *iter ), QItemSelectionModel::Select|QItemSelectionModel::Rows ); }

    return;

}

//_________________________________________________________
void GeometryDialog::_reorder( int oldRow, int newRow )
{

    Debug::Throw() << "GeometryDialog::_reorder - old: " << oldRow << " new: " << newRow << endl;
    if( newRow > oldRow ) newRow--;

    ApplicationDataModel::List dataList( model_.get() );
    dataList.move( oldRow, newRow );
    model_.set( dataList );

    // restore selection
    list_->selectionModel()->select( model_.index( newRow, 0 ),  QItemSelectionModel::Clear|QItemSelectionModel::Select|QItemSelectionModel::Rows );
    return;

}

//____________________________________________________
void GeometryDialog::_apply( bool warn )
{

    Debug::Throw( "GeometryDialog::_apply.\n" );
    ApplicationDataModel::List applications( model_.get() );

    // store order
    int index(0);
    for( ApplicationDataModel::List::iterator iter = applications.begin(); iter != applications.end(); ++iter, ++index )
    { iter->setOrder( index ); }

    // save backup
    if( backup_.empty() ) backup_ = applications;

    // get configuration
    const int alignmentIndex( alignmentComboBox_->currentIndex() );
    const Anchor::AlignmentFlags anchor = anchorWidget_->anchor();
    const int spacing = spacingSpinBox_->value();
    const int offset = offsetSpinBox_->value();

    // get desktop size
    const QRect desktop( QApplication::desktop()->screenGeometry( QApplication::desktop()->screenNumber( this ) ) );

    if( alignmentIndex == 1 )
    {

        // horizontal alignment
        int xBase(0);
        if( anchor&Anchor::Left ) xBase = desktop.left() + spacing;
        else if( anchor&Anchor::Right ) xBase = desktop.right() - spacing;
        else xBase = (desktop.left()+desktop.right());

        // vertical alignment
        int yBase(0);
        if( anchor&Anchor::Top ) yBase = desktop.top() + offset;
        else if( anchor&Anchor::VerticalCenter )
        {

            // calculate full height
            int height(0);
            for( ApplicationDataModel::List::iterator iter = applications.begin(); iter != applications.end(); ++iter )
            { height += iter->geometry().height() + spacing; }
            height -= spacing;

            // calculate offset
            yBase = desktop.top() + ( desktop.height() - height )/2;


        }

        // top and center alignment
        if( anchor & (Anchor::Top|Anchor::VerticalCenter) )
        {

            // top and center alignment
            for( ApplicationDataModel::List::iterator iter = applications.begin(); iter != applications.end(); ++iter )
            {
                int x( xBase );
                if( anchor&Anchor::Right ) x -= iter->geometry().width();
                else if( anchor&Anchor::HorizontalCenter ) {
                    x -= iter->geometry().width();
                    x /= 2;
                }

                iter->setGeometry( QRect( QPoint( x, yBase ), iter->geometry().size() ) );
                yBase += iter->geometry().height() + spacing;

            }

        } else {

            // bottom alignment
            int yBase( desktop.bottom() - offset );
            ApplicationDataModel::MutableListIterator iter( applications );
            iter.toBack();
            while( iter.hasPrevious() )
            {
                ApplicationData& current( iter.previous() );
                yBase -= current.geometry().height();
                int x( xBase );
                if( anchor&Anchor::Right ) x -= current.geometry().width();
                else if( anchor&Anchor::HorizontalCenter ) {
                    x -= current.geometry().width();
                    x /= 2;
                }

                current.setGeometry( QRect( QPoint( x, yBase ), current.geometry().size() ) );
                yBase -= spacing;
            }

        }

    } else {

        // horizontal alignment
        int yBase(0);
        if( anchor&Anchor::Top ) yBase = desktop.top() + spacing;
        else if( anchor&Anchor::Bottom ) yBase = desktop.bottom() - spacing;
        else yBase = (desktop.top()+ desktop.bottom());

        // vertical alignment
        int xBase(0);
        if( anchor&Anchor::Left ) xBase = desktop.left() + offset;
        else if( anchor&Anchor::HorizontalCenter )
        {

            // calculate full height
            int width(0);
            for( ApplicationDataModel::List::iterator iter = applications.begin(); iter != applications.end(); ++iter )
            { width += iter->geometry().width() + spacing; }
            width -= spacing;

            // calculate offset
            xBase = desktop.left() + ( desktop.width() - width )/2;


        }

        if( anchor & (Anchor::Left|Anchor::HorizontalCenter) )
        {

            // left and center alignment
            for( ApplicationDataModel::List::iterator iter = applications.begin(); iter != applications.end(); ++iter )
            {

                int y( yBase );
                if( anchor&Anchor::Bottom ) y -= iter->geometry().height();
                else if( anchor&Anchor::VerticalCenter ) {
                    y -= iter->geometry().height();
                    y /= 2;
                }

                iter->setGeometry( QRect( QPoint( xBase, y ), iter->geometry().size() ) );
                xBase += iter->geometry().width() + spacing;

            }

        } else {

            // right alignment
            int xBase( desktop.right() - offset );
           ApplicationDataModel::MutableListIterator iter( applications );
            iter.toBack();
            while( iter.hasPrevious() )
            {
                ApplicationData& current( iter.previous() );
                xBase -= current.geometry().width();
                int y( yBase );
                if( anchor&Anchor::Bottom ) y -= current.geometry().height();
                else if( anchor&Anchor::VerticalCenter ) {
                    y -= current.geometry().height();
                    y /= 2;
                }

                current.setGeometry( QRect( QPoint( xBase, y ), current.geometry().size() ) );
                xBase -= spacing;

            }

        }

    }

    // get model current set
    ApplicationDataModel::List current( model_.get() );
    ApplicationDataModel::List newApplications;

    // update model
    model_.set( applications );

    // send new geometries to client
    foreach( const ApplicationData& data, applications )
    {

        ApplicationDataModel::List::iterator currentIter( std::find( current.begin(), current.end(), data ) );
        if( currentIter != current.end() && currentIter->geometry() != data.geometry() )
        { newApplications.append( data ); }

    }

    // check, ask confirmation and emit signals
    if( newApplications.empty() )
    {
        if( warn )
        { InformationDialog( this, tr( "There are no changes to apply with respect to current geometry. <Apply> Canceled." ) ).exec(); }
        return;
    }
    if( !GeometryConfirmationDialog( this, newApplications ).exec() ) return;
    foreach( const ApplicationData& data, newApplications )
    { emit geometryChanged( data ); }

}

//____________________________________________________
void GeometryDialog::_restoreDefaults( void )
{

    if( backup_.empty() ) return;
    ApplicationDataModel::List current( model_.get() );
    for( ApplicationDataModel::List::iterator iter = backup_.begin(); iter != backup_.end(); ++iter )
    {

        ApplicationDataModel::List::iterator currentIter( std::find( current.begin(), current.end(), *iter ) );
        if( currentIter != current.end() && currentIter->geometry() != iter->geometry() )
        { emit geometryChanged( *iter ); }

    }

}

//____________________________________________________
void GeometryDialog::_updateConfiguration( void )
{

    Debug::Throw( "GeometryDialog::_updateConfiguration.\n" );
    OptionWidgetList::read( XmlOptions::get() );

}

//____________________________________________________
void GeometryDialog::_saveConfiguration( void )
{

    Debug::Throw( "GeometryDialog::_saveConfiguration.\n" );
    OptionWidgetList::write( XmlOptions::get() );

}
