// $Id: PictureListDialog.cpp,v 1.24 2014/09/26 08:06:19 hupereir Exp $

/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "PictureListDialog.h"
#include "PictureListDialog.moc"

#include "TreeView.h"
#include "EditPictureDialog.h"
#include "QtUtil.h"
#include "Singleton.h"

#include <QHeaderView>

namespace Pictures
{

    //________________________________________________________
    PictureListDialog::PictureListDialog( QWidget* parent, const PictureAttributes::List& list ):
        CustomDialog( parent )
    {

        Debug::Throw( "PictureListDialog::PictureListDialog.\n" );

        setWindowTitle( tr( "Picture Viewer" ) );

        QHBoxLayout* layout( new QHBoxLayout() );
        layout->setSpacing(5);
        layout->setMargin(0);
        mainLayout().addLayout( layout );

        // create list
        list_ = new TreeView( this );
        _list().setModel( &model_ );
        _list().header()->hide();
        _list().setSortingEnabled( false );
        layout->addWidget( &_list(), 1 );

        // set connections
        connect( &_list(), SIGNAL(clicked(QModelIndex)), SLOT(_toggle(QModelIndex)) );
        connect( &_list(), SIGNAL(activated(QModelIndex)), SLOT(_edit()) );
        connect( _list().selectionModel(), SIGNAL(selectionChanged(QItemSelection,QItemSelection)), SLOT(_updateButtons()) );

        connect( Singleton::get().application(), SIGNAL(configurationChanged()), SLOT(_updateConfiguration()) );

        // buttons
        QVBoxLayout* button_layout = new QVBoxLayout();
        button_layout->setMargin(0);
        button_layout->setSpacing( 5 );
        layout->addLayout( button_layout, 0 );

        // Add button
        QPushButton *button;
        button_layout->addWidget( button = new QPushButton( tr( "Add" ), this ) );
        connect( button, SIGNAL(clicked()), SLOT(_add()) );

        // Edit button
        button_layout->addWidget( edit_ = new QPushButton( tr( "Edit" ), this ) );
        connect( edit_, SIGNAL(clicked()), SLOT(_edit()) );
        edit_->setToolTip( tr( "Edit selected pictures" ) );

        // remove action
        addAction( removeAction_ = new QAction( tr( "Remove" ), this ) );
        connect( removeAction_, SIGNAL(triggered()), SLOT(_remove()) );
        removeAction_->setShortcut( QKeySequence::Delete );
        removeAction_->setToolTip( tr( "Remove selected pictures" ) );

        // remove button
        button_layout->addWidget( remove_ = new QPushButton( tr( "Remove" ), this ) );
        connect( remove_, SIGNAL(clicked()), SLOT(_remove()) );
        remove_->setToolTip( tr( "Remove selected pictures" ) );

        QFrame* frame = new QFrame( this );
        frame->setFrameStyle( QFrame::HLine | QFrame::Plain );
        button_layout->addWidget( frame );

        button_layout->addWidget( selectAllButton_ = new QPushButton( tr( "Select All" ), this ) );
        connect( selectAllButton_, SIGNAL(pressed()), SLOT(_selectAll()) );

        button_layout->addWidget( selectNoneButton_ = new QPushButton( tr( "Select None" ), this ) );
        connect( selectNoneButton_, SIGNAL(pressed()), SLOT(_selectNone()) );

        button_layout->addWidget( invertButton_ = new QPushButton( tr( "Invert Selection" ), this ) );
        connect( invertButton_, SIGNAL(pressed()), SLOT(_invert()) );

        button_layout->addStretch(1);

        // update configuration
        _updateConfiguration();

        // set list
        setAttributes( list );

        // update buttons
        _updateButtons();

    }

    //_____________________________________________________________________
    void PictureListDialog::setAttributes( const PictureAttributes::List& list )
    {

        Debug::Throw( "PictureListDialog::setList.\n" );
        _model().update( list );
        _list().resizeColumns();
    }

    //_____________________________________________________________________
    PictureAttributes::List PictureListDialog::attributes( void ) const
    {

        Debug::Throw( "PictureListDialog::list.\n" );
        return _model().get();
    }

    //_________________________________________________________________________
    void PictureListDialog::_updateConfiguration( void )
    {
        Debug::Throw( "PictureListDialog::_updateConfiguration.\n" );
        PictureAttributesModel::clearIconCache();
        _model().set( _model().get() );
    }

    //_________________________________________________________________________
    void PictureListDialog::_updateButtons( void )
    {
        Debug::Throw( "PictureListDialog::_updateButtons.\n" );
        int selection_size( _list().selectionModel()->selectedRows().size() );

        // enable buttons depending on the size of the list
        edit_->setEnabled( selection_size == 1 );
        removeAction_->setEnabled( selection_size != 0 );
        remove_->setEnabled( selection_size != 0 );

        bool has_checked( false );
        bool has_unchecked( false );

        // loop over files
        PictureAttributesModel::List attributes( _model().get() );
        for( PictureAttributesModel::List::const_iterator iter = attributes.begin(); iter != attributes.end(); ++iter )
        {

            if( iter->isVisible() ) has_checked = true;
            else has_unchecked = true;

        }

        selectAllButton_->setEnabled( has_unchecked );
        selectNoneButton_->setEnabled( has_checked );
        invertButton_->setEnabled( true );

    }

    //_________________________________________________________________________
    void PictureListDialog::_add( void )
    {
        Debug::Throw( "PictureListDialog::_add.\n" );

        EditPictureDialog dialog( this );
        dialog.setAttributes( PictureAttributes() );
        if( dialog.centerOnWidget( window() ).exec() == QDialog::Rejected ) return;

        // add item
        _model().add( dialog.attributes() );
        _list().resizeColumns();
        return;
    }

    //_________________________________________________________________________
    void PictureListDialog::_edit( void )
    {
        Debug::Throw( "PictureListDialog::_edit.\n" );

        // retrieve selection
        QModelIndex current( _list().selectionModel()->currentIndex() );
        if( !current.isValid() ) return;

        EditPictureDialog dialog( this );
        dialog.setAttributes( _model().get( current ) );
        if( dialog.centerOnWidget( window() ).exec() == QDialog::Rejected ) return;

        _model().replace( current, dialog.attributes() );
        _list().resizeColumns();

        return;

    }

    //__________________________________________________
    void PictureListDialog::_remove( void )
    {
        Debug::Throw( "PictureListDialog::_remove.\n" );

        // retrieve selected items; make sure they do not include the navigator
        _model().remove( _model().get( _list().selectionModel()->selectedRows() ) );
        _list().resizeColumns();

    }

    //_____________________________________________________________
    void PictureListDialog::_selectAll(void)
    {
        Debug::Throw( "PictureListDialog::_selectAll.\n" );

        // get all files
        PictureAttributesModel::List attributes( _model().get() );

        // loop over files
        for( PictureAttributesModel::List::const_iterator iter = attributes.begin(); iter != attributes.end(); ++iter )
        {
            if( !iter->isVisible() ) _toggle( _model().index( *iter, PictureAttributesModel::Visibility ) );
        }

        selectAllButton_->setEnabled( false );
        selectNoneButton_->setEnabled( true );

    }

    //_____________________________________________________________
    void PictureListDialog::_selectNone(void)
    {
        Debug::Throw( "PictureListDialog::_selectNone.\n" );

        // get all files
        PictureAttributesModel::List attributes( _model().get() );

        // loop over files
        for( PictureAttributesModel::List::const_iterator iter = attributes.begin(); iter != attributes.end(); ++iter )
        {
            if( iter->isVisible() ) _toggle( _model().index( *iter, PictureAttributesModel::Visibility ) );
        }

        selectAllButton_->setEnabled( true );
        selectNoneButton_->setEnabled( false );

    }

    //_____________________________________________________________
    void PictureListDialog::_invert(void)
    {
        Debug::Throw( "PictureListDialog::_invert.\n" );

        // get all files
        PictureAttributesModel::List attributes( _model().get() );

        // loop over files
        for( PictureAttributesModel::List::const_iterator iter = attributes.begin(); iter != attributes.end(); ++iter )
        {
            _toggle( _model().index( *iter, PictureAttributesModel::Visibility ) );
        }

    }

    //_______________________________________________
    void PictureListDialog::_toggle( const QModelIndex& index )
    {
        Debug::Throw( "PictureListDialog::_toggle.\n" );

        if( !index.isValid() ) return;
        if( index.column() != PictureAttributesModel::Visibility ) return;

        PictureAttributes attributes( _model().get( index ) );
        attributes.setVisible( !attributes.isVisible() );
        _model().replace( index, attributes );

        return;
    }

}
